# -*- coding: utf-8 -*-
from pkg_resources import resource_filename, resource_listdir

import click
from path import path


@click.command()
@click.argument('location', type=click.Path(), default='./chanjo-demo',
                required=False)
@click.pass_context
def demo(context, location):
    """Copy demo files to a directory.

    \b
    LOCATION: directory to add demofiles to (default: ./chanjo-demo)
    """
    user_dir = path(location)
    pkg_dir = __name__.rpartition('.')[0]
    demo_dir = path(resource_filename(pkg_dir, 'files'))

    # make sure we don't overwrite exiting files
    for demo_file in resource_listdir(pkg_dir, 'files'):
        user_file_path = user_dir.joinpath(demo_file)
        if user_file_path.exists():
            click.echo("{} exists, pick a different location"
                       .format(user_file_path))
            context.abort()

    try:
        # we can copy the directory(tree)
        demo_dir.copytree(user_dir)
    except OSError:
        click.echo('The location must be a non-existing directory.')
        context.abort()

    # inform the user
    click.echo("Successfully copied demo files to {}".format(user_dir))



# -*- coding: utf-8 -*-
from chanjo.store import Store


class TestStore(object):
    def setup(self):
        self.store = Store('sqlite://')
        self.store.set_up()

    def test_init(self):
        pass


def test_lazy_load():
    """Test connection after post-init."""
    store = Store()

    # now we shouldn't have access to aliases
    assert hasattr(store, 'query') == False

    store.connect('sqlite://')

    # ... but now we do!
    assert hasattr(store, 'query')
    assert store.dialect == 'sqlite'



# -*- coding: utf-8 -*-
from __future__ import division
import itertools
import logging

from sqlalchemy.sql import func

from chanjo.compat import itervalues
from chanjo.utils import list_get
from .core import Store
from .converter import ChanjoConverterMixin
from .models import Exon, ExonStatistic, Sample, Transcript
from .utils import filter_samples, group_by_field, predict_gender

logger = logging.getLogger(__name__)


class ChanjoAPI(Store, ChanjoConverterMixin):

    """Interface to chanjo database for asking interesting questions.

    Attributes:
        weighten_average (BinaryExpression): weighter mean for metrics
    """

    def init_app(self, app, key_base='CHANJO_'):
        """Configure API (Flask style) after lazy initialization.

        Args:
            app (Flask): Flask app instance
            key_base (str): namespace to look for under ``app.config``

        Returns:
            ChanjoAPI: ``self``
        """
        uri = app.config["{}URI".format(key_base)]
        self.connect(db_uri=uri)
        return self

    @property
    def weighted_average(self):
        """Build a dynamic field to calculate a weighted average.

        Works for any of the exon metrics.

        Returns:
            BinaryExpression: dynamic field for a query
        """
        # length of the exon
        weight = Exon.end - Exon.start
        total_weight = func.sum(weight)
        total_value = func.sum(ExonStatistic.value * weight)
        weighted_mean = (total_value / total_weight).label('weighted_mean')
        return weighted_mean

    def samples(self, group_id=None, sample_ids=None):
        """Get samples from the database.

        Args:
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            List[Sample]: samples objects from the database
        """
        query = self.query(Sample).order_by(Sample.sample_id)
        query = filter_samples(query, group_id=group_id, sample_ids=sample_ids)
        return query

    def means(self, query=None):
        """Calculate means on a per sample basis.

        Args:
            query (Optional[Query]): initialized and filtered query

        Returns:
            List[tuple]: sample id and grouped mean metrics
        """
        query = query or self.query()
        ready_query = (query.add_columns(Sample.sample_id,
                                         ExonStatistic.metric,
                                         self.weighted_average)
                            .join(ExonStatistic.sample, ExonStatistic.exon)
                            .group_by(Sample.sample_id, ExonStatistic.metric)
                            .order_by(Sample.sample_id))
        results = group_by_field(ready_query)
        return results

    def region_alt(self, region_id, sample_id=None, per=None):
        """Parse region id as input for `region` method.

        Args:
            region_id (str): for example 1:1242314-1243419
            sample_id (Optional[str]): filter to only a single sample
            per (Optional[str]): report on a per exon basis ("exon")

        Returns:
            dict: weighted metrics (across samples)
        """
        chromosome, pos_str = region_id.split(':')
        start, end = [int(pos) for pos in pos_str.split('-')]
        data = self.region(chromosome, start, end, sample_id=sample_id,
                           per=per)
        return data

    def region(self, chromosome, start, end, sample_id=None, per=None):
        """Report coverage across a genomics region (of exons).

        Args:
            chromosome (str): chromosome id
            start (int): start position (inclusive)
            end (int): end position (inclusive)
            sample_id (Optional[str]): filter to only a single sample
            per (Optional[str]): report on a per exon basis ("exon")

        Returns:
            dict: weighted metrics (across samples)
        """
        results = (self.query(Exon.exon_id, ExonStatistic.metric,
                              self.weighted_average)
                       .join(ExonStatistic.exon)
                       .filter(Exon.chromosome == chromosome,
                               Exon.start >= start,
                               Exon.end <= end)
                       .group_by(ExonStatistic.metric)
                       .order_by(Exon.exon_id))

        if sample_id:
            results = (results.join(ExonStatistic.sample)
                              .filter(Sample.sample_id == sample_id))

        if per == 'exon':
            results = results.group_by(Exon.exon_id)
            exon_groups = group_by_field(results)
            return exon_groups
        else:
            data = {metric: value for exon_id, metric, value in results}
            return data

    def gene(self, *gene_ids):
        r"""Report aggregate statistics for particular genes.

        Args:
            \*gene_ids (List[str]): gene ids

        Returns:
            List[dict]: metrics grouped by sample and gene
        """
        samples = {}
        for gene_id in gene_ids:
            logger.debug('figure out which transcripts the gene belongs to')
            exon_objs = self.gene_exons(gene_id).all()
            if len(exon_objs) == 0:
                raise AttributeError("gene id not in database: {}"
                                     .format(gene_id))
            exon_ids = [exon_obj.exon_id for exon_obj in exon_objs]
            query = (self.query().filter(Exon.exon_id.in_(exon_ids)))
            for sample_id, data in self.means(query):
                if sample_id not in samples:
                    samples[sample_id] = {'sample_id': sample_id, 'genes': {}}
                samples[sample_id]['genes'][gene_id] = data

        return itervalues(samples)

    def transcripts(self, *transcript_ids):
        r"""Report metrics on specific transcripts.

        Args:
            \*transcript_ids (List[str]): transcript ids

        Returns:
            List[tuple]: sample, transcript, metric, value
        """
        results = (self.query(Sample.sample_id,
                              Transcript.transcript_id,
                              ExonStatistic.metric,
                              self.weighted_average)
                       .join(ExonStatistic.sample, ExonStatistic.exon,
                             Exon.transcripts)
                       .filter(Transcript.transcript_id.in_(transcript_ids))
                       .group_by(Sample.sample_id, ExonStatistic.metric,
                                 Transcript.transcript_id))
        return results

    def incomplete_exons(self, query=None, level=10, threshold=100):
        """Fetch exons with incomplete coverage at a specifc level.

        Args:
            level (Optional[int]): completeness level, default: 10
            threshold (Optional[int]): % completeness cutoff, default: 100
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            List[tuple]: sample_id, dict with exons and completeness
        """
        completeness = "completeness_{}".format(level)
        query = query or self.query()
        ready_query = (self.query(Sample.sample_id,
                                  Exon.exon_id,
                                  ExonStatistic.value)
                           .join(ExonStatistic.exon, ExonStatistic.sample)
                           .filter(ExonStatistic.metric == completeness,
                                   ExonStatistic.value < threshold)
                           .order_by(Sample.sample_id))
        exon_samples = group_by_field(ready_query)
        return exon_samples

    def sex_coverage(self, sex_chromosomes=('X', 'Y')):
        """Query for average on X/Y chromsomes.

        Args:
            sex_chromosomes (Optional[tuple]): chromosome ids

        Returns:
            List[tuple]: sample, chromosome, weighted average coverage
        """
        query = (self.query(Sample.sample_id,
                            Exon.chromosome,
                            self.weighted_average)
                     .join(ExonStatistic.exon, ExonStatistic.sample)
                     .filter(Exon.chromosome.in_(sex_chromosomes),
                             ExonStatistic.metric == 'mean_coverage')
                     .group_by(Sample.sample_id, Exon.chromosome))
        return query

    def sex_check(self, group_id=None, sample_ids=None):
        """Predict gender based on coverage of sex chromosomes.

        Args:
            group_id (Optional[str]): sample group identifier
            sample_ids (Optional[List[str]]): sample ids

        Returns:
            tuple: sample, gender, coverage X, coverage, Y
        """
        query = self.sex_coverage()
        query = filter_samples(query, group_id=group_id, sample_ids=sample_ids)
        logger.debug('group rows based on sample')
        samples = itertools.groupby(query, lambda row: row[0])
        for sample_id, chromosomes in samples:
            sex_coverage = [coverage for _, _, coverage in chromosomes]
            logger.debug('predict gender')
            # run the predictor
            gender = predict_gender(*sex_coverage)
            yield sample_id, gender, sex_coverage[0], sex_coverage[1]

    def diagnostic_yield(self, sample_id, query=None, level=10, threshold=100):
        """Calculate transcripts that aren't completely covered."""
        query = query or self.query()
        level = "completeness_{}".format(level)

        all_query = (query.add_columns(Transcript.transcript_id)
                          .distinct(Transcript.transcript_id)
                          .join(Exon.transcripts))
        all_count = all_query.count()

        logger.debug('find out which exons failed')
        yield_query = (query.add_columns(Exon.exon_id)
                            .join(ExonStatistic.exon)
                            .filter(ExonStatistic.metric == level,
                                    ExonStatistic.value < threshold))
        exon_ids = [row[0] for row in yield_query]
        transcripts = self.exon_transcripts(exon_ids)

        tx_count = transcripts.count()
        diagnostic_yield = 100 - (tx_count/all_count * 100)
        return {
            "diagnostic_yield": diagnostic_yield,
            "count": tx_count,
            "total_count": all_count,
            "transcripts": transcripts
        }

    def completeness_levels(self):
        """Return a list of included completeness levels."""
        metrics = (self.query(ExonStatistic.metric)
                       .distinct(ExonStatistic.metric))
        # for all completeness levels, extract the level as int and full name
        levels = ((int(field[0].split('_')[-1]), field[0])
                  for field in metrics
                  if field[0].startswith('completeness'))
        # sort them based on the level int
        sorted_levels = sorted(levels, key=lambda level: list_get(level, 0))
        return sorted_levels



# -*- coding: utf-8 -*-
import logging

import click

from chanjo.load import link as link_mod
from chanjo.parse import bed
from chanjo.store import Store
from chanjo.utils import validate_stdin

logger = logging.getLogger(__name__)


@click.command()
@click.argument('bed_stream', callback=validate_stdin,
                type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def link(context, bed_stream):
    """Link related genomic elements."""
    chanjo_db = Store(uri=context.obj['database'])
    link_elements(chanjo_db, bed_stream)


def link_elements(chanjo_db, bed_iterable, batch_size=10000):
    """Load Sambamba BED output from a stream."""
    rows = bed.chanjo(bed_iterable)
    stats = link_mod.rows(chanjo_db.session, rows)
    for index, stat in enumerate(stats):
        chanjo_db.add(stat)
        if index % batch_size == 0:
            chanjo_db.save()
            logger.debug('processed %s exons...', index)

    chanjo_db.save()



# -*- coding: utf-8 -*-
import json

import click


def dump_json(*items):
    """Dump list of dicts as json objects.

    Args:
        items (List[dict]): list of JSON serializable dicts
    """
    for item in items:
        json_dump = json.dumps(item)
        click.echo(json_dump)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
chanjo.__main__
~~~~~~~~~~~~~~~~

The main entry point for the command line interface.

Invoke as ``chanjo`` (if installed) or ``python -m chanjo`` (no install
required).
"""
import sys

from .cli import root_command


if __name__ == '__main__':
    # exit using whatever exit code the main returned
    sys.exit(root_command())



# -*- coding: utf-8 -*-
from datetime import datetime

from sqlalchemy import (Column, DateTime, Float, ForeignKey, Integer, String,
                        Table, UniqueConstraint)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declarative_base

# base for declaring a mapping
BASE = declarative_base()


# +--------------------------------------------------------------------+
# | Association tables
# | ~~~~~~~~~~~~~~~~~~~
# | Provides the many-to-many relationships between:
# | - Exon<->Transcript
# +--------------------------------------------------------------------+
Exon_Transcript = Table(
    'exon__transcript',
    BASE.metadata,
    Column('exon_id', Integer, ForeignKey('exon.id')),
    Column('transcript_id', Integer, ForeignKey('transcript.id')))


# +--------------------------------------------------------------------+
# | Gene ORM
# +--------------------------------------------------------------------+
class Gene(BASE):

    """Collection of transcripts and potentially overlapping exons.

    A :class:`Gene` can be related to multiple transcripts and multiple
    exons.

    Args:
        gene_id (str): unique gene id e.g. HGNC gene symbol
    """

    __tablename__ = 'gene'
    __table_args__ = (UniqueConstraint('chromosome', 'start', 'end',
                                       name='_coordinates'),)

    id = Column(Integer, primary_key=True)
    gene_id = Column(String(32), unique=True)
    chromosome = Column(String(32))
    start = Column(Integer)
    end = Column(Integer)


# +--------------------------------------------------------------------+
# | Transcript ORM
# +--------------------------------------------------------------------+
class Transcript(BASE):

    """Set of non-overlapping exons.

    A :class:`Transcript` can *only* be related to a single gene.

    Args:
        transcript_id (str): unique block id (e.g. CCDS transcript id)
        gene_id (str): related gene
    """

    __tablename__ = 'transcript'

    id = Column(Integer, primary_key=True)
    transcript_id = Column(String(32))

    gene_id = Column(String(32), ForeignKey('gene.id'))
    gene = relationship(Gene, backref=backref('transcripts'))


# +--------------------------------------------------------------------+
# | Exon ORM
# +--------------------------------------------------------------------+
class Exon(BASE):

    """A continous genetic interval on a given contig.

    A :class:`Exon` can be related to a multiple :class:`Transcript`.
    Start and end coordinates are 1-based.

    Args:
        exon_id (str): unique exon id
        contig (str): contig/chromosome id
        start (int): 1-based start of the exon
        end (int): 1-based end of the exon
    """

    __tablename__ = 'exon'
    __table_args__ = (UniqueConstraint('chromosome', 'start', 'end',
                                       name='_coordinates'),)

    id = Column(Integer, primary_key=True)
    exon_id = Column(String(32), unique=True)
    chromosome = Column(String(32))
    start = Column(Integer)
    end = Column(Integer)

    transcripts = relationship(Transcript, secondary=Exon_Transcript,
                               backref=backref('exons', order_by=start))

    def __len__(self):
        """Return the number of bases.

        Returns:
            int: length of interval in number of bases
        """
        # add +1 because both coordinates are 1-based (number of *bases*)
        return (self.end - self.start) + 1


# +--------------------------------------------------------------------+
# | Sample ORM classes
# +--------------------------------------------------------------------+
class Sample(BASE):

    """Metadata for a single (unique) sample.

    :class:`Sample` helps out in consolidating important information in
    one place.

    .. versionadded:: 0.4.0

    Args:
        sample_id (str): unique sample id
        group_id (str): unique group id
        cutoff (int): cutoff used for completeness
        extension (bool): number of bases added to each interval
        coverage_source (str): path to the BAM file used
    """

    __tablename__ = 'sample'

    id = Column(Integer, primary_key=True)
    sample_id = Column(String(32), unique=True)
    group_id = Column(String(32), index=True)
    created_at = Column(DateTime, default=datetime.now)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)


# +--------------------------------------------------------------------+
# | Exon Data ORM
# +--------------------------------------------------------------------+
class Statistic(object):

    """Coverage metrics for a single element and a given sample.

    Args:
        metric (str): identifier for the metric
        value (float): value for the metric
    """

    id = Column(Integer, primary_key=True)
    metric = Column(String(32))
    value = Column(Float)


class ExonStatistic(Statistic, BASE):

    """Statistics on the exon level, related to sample and exon.

    Args:
        sample (Sample): parent record Sample
        exon (Exon): parent record Exon
        sample_id (int): parent record Sample id
        exon_id (int): parent record Exon id
    """

    __tablename__ = 'exon_stat'

    sample_id = Column(Integer, ForeignKey('sample.id'))
    sample = relationship(Sample, backref=backref('exon_stats'))
    exon_id = Column(Integer, ForeignKey('exon.id'))
    exon = relationship(Exon, backref=backref('stats'))



# -*- coding: utf-8 -*-
from chanjo.store import Gene, Transcript

from .utils import get_or_build_exon


def rows(session, row_data):
    """Handle rows of sambamba output."""
    exons = (row(session, data) for data in row_data)
    return exons


def row(session, data):
    """Link transcripts and genes."""
    # start with exons object
    exon_obj = get_or_build_exon(session, data)
    # store created gene objects
    genes = {}
    for tx_id, gene_id in data['elements']:
        gene_obj = session.query(Gene).filter_by(gene_id=gene_id).first()
        if gene_obj is None:
            # create new gene and store to avoid conflicts
            genes[gene_id] = gene_obj = (genes.get(gene_id) or
                                         Gene(gene_id=gene_id))

        tx_filters = {'transcript_id': tx_id}
        tx_obj = session.query(Transcript).filter_by(**tx_filters).first()
        if tx_obj is None:
            # create new transcript and link with gene
            tx_obj = Transcript(**tx_filters)
            tx_obj.gene = gene_obj

        if tx_obj not in exon_obj.transcripts:
            # link exon to the transcript
            exon_obj.transcripts.append(tx_obj)

    return exon_obj



# -*- coding: utf-8 -*-
"""
chanjo.store.core
~~~~~~~~~~~~~~~~~~
"""
from __future__ import division
import os

from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql.expression import ClauseElement

from .models import (BASE, Gene, Transcript, Exon, Sample)


class Store(object):

    """SQLAlchemy-based database object.

    Bundles functionality required to setup and interact with various
    related genomic interval elements.

    .. versionchanged:: 2.1.0
        Lazy-loadable, all "init" arguments optional.

    Examples:
        >>> chanjo_db = Store('data/elements.sqlite3')
        >>> chanjo_db.set_up()

    .. note::

        For testing pourposes use ``:memory:`` as the ``path`` argument to
        set up in-memory (temporary) database.

    Args:
        uri (Optional[str]): path/URI to the database to connect to
        debug (Optional[bool]): whether to output logging information

    Attributes:
        uri (str): path/URI to the database to connect to
        engine (class): SQLAlchemy engine, defines what database to use
        session (class): SQLAlchemy ORM session, manages persistance
        query (method): SQLAlchemy ORM query builder method
        classes (dict): bound ORM classes
    """

    def __init__(self, uri=None, debug=False):
        super(Store, self).__init__()
        self.uri = uri
        if uri:
            self.connect(uri, debug=debug)

        # ORM class shortcuts to enable fetching models dynamically
        self.classes = {'gene': Gene, 'transcript': Transcript,
                        'exon': Exon, 'sample': Sample}

    def connect(self, db_uri, debug=False):
        """Configure connection to a SQL database.

        .. versionadded:: 2.1.0

        Args:
            db_uri (str): path/URI to the database to connect to
            debug (Optional[bool]): whether to output logging information
        """
        kwargs = {'echo': debug, 'convert_unicode': True}
        # connect to the SQL database
        if 'mysql' in db_uri:
            kwargs['pool_recycle'] = 3600
        elif '://' not in db_uri:
            # expect only a path to a sqlite database
            db_path = os.path.abspath(os.path.expanduser(db_uri))
            db_uri = "sqlite:///{}".format(db_path)

        self.engine = create_engine(db_uri, **kwargs)
        # make sure the same engine is propagated to the BASE classes
        BASE.metadata.bind = self.engine
        # start a session
        self.session = scoped_session(sessionmaker(bind=self.engine))
        # shortcut to query method
        self.query = self.session.query
        return self

    @property
    def dialect(self):
        """Return database dialect name used for the current connection.

        Dynamic attribute.

        Returns:
            str: name of dialect used for database connection
        """
        return self.engine.dialect.name

    def set_up(self):
        """Initialize a new database with the default tables and columns.

        Returns:
            Store: self
        """
        # create the tables
        BASE.metadata.create_all(self.engine)
        return self

    def tear_down(self):
        """Tear down a database (tables and columns).

        Returns:
            Store: self
        """
        # drop/delete the tables
        BASE.metadata.drop_all(self.engine)
        return self

    def get_or_create(self, model, **kwargs):
        """Get or create a record in the database."""
        try:
            query = self.query(model).filter_by(**kwargs)
            instance = query.first()
            if instance:
                return instance, False
            else:
                self.session.begin(nested=True)
                try:
                    params = dict((key, value) for key, value
                                  in kwargs.iteritems()
                                  if not isinstance(value, ClauseElement))
                    instance = model(**params)
                    self.session.add(instance)
                    self.session.commit()
                    return instance, True
                except IntegrityError as error:
                    self.session.rollback()
                    instance = query.one()
                    return instance, False
        except Exception as exception:
            raise exception

    def save(self):
        """Manually persist changes made to various elements. Chainable.

        .. versionchanged:: 2.1.2
            Flush session before commit.

        Returns:
            Store: ``self`` for chainability
        """
        # commit/persist dirty changes to the database
        self.session.flush()
        self.session.commit()
        return self

    def get(self, typ, type_id):
        """Fetch a specific element or ORM class.

        Calls itself recursively when asked to fetch an element.

        Args:
            typ (str): element key or 'class'
            type_id (str): element id or ORM model id

        Returns:
            model: element or ORM class

        Examples:
            >>> gene = db.get('gene', 'GIT1')
        """
        if typ == 'class':
            return self.classes.get(type_id, None)

        # get an ORM class (recursive)
        klass = self.get('class', typ)

        # return the requested element object (or ``None``) if not found
        return self.session.query(klass).get(type_id)

    def find(self, klass_id, query=None, attrs=None):
        """Fetch one or more elements based on the query.

        If the 'query' parameter is a string :meth:`~chanjo.Store.find`
        will fetch one element; just like `get`. If query is a list it will
        match element ids to items in that list and return a list of
        elements. If 'query' is ``None`` all elements of that class will
        be returned.

        Args:
            klass_id (str): type of element to find
            query (str/list, optional): element Id(s)
            attrs (list, optional): list of columns to fetch

        Returns:
            object/list: element(s) from the database
        """
        # get an ORM class
        klass = self.get('class', klass_id)

        if attrs is not None:
            params = [getattr(klass, attr) for attr in attrs]
        else:
            params = (klass,)

        if query is None:
            # return all `klass_id` elements in the database
            return self.query(*params).all()
        elif isinstance(query, list):
            # return all `klass_id` elements in the database
            return self.query(*params).filter(klass.id.in_(query)).all()
        elif isinstance(query, str):
            # call 'get' to return the single element
            return self.get(klass_id, query)
        else:
            raise ValueError("'query' must be 'None', 'list', or 'str'")

    def add(self, elements):
        """Add one or more new elements and commit the changes. Chainable.

        Args:
            elements (orm/list): new ORM object instance or list of such

        Returns:
            Store: ``self`` for chainability
        """
        if isinstance(elements, BASE):
            # Add the record to the session object
            self.session.add(elements)
        elif isinstance(elements, list):
            # Add all records to the session object
            self.session.add_all(elements)
        return self

    def create(self, class_id, *args, **kwargs):
        r"""Create a new instance of an ORM element.

        If attributes are supplied as a tuple they must be in the correct
        order. Supplying a `dict` doesn't require the attributes to be in
        any particular order.

        Args:
            class_id (str): choice between "superblock", "block", "interval"
            \*args (tuple): list the element attributes in the *correct order*
            \**kwargs (dict): element attributes in whatever order you like

        Returns:
            orm: new ORM instance object
        """
        if args:
            # unpack tuple
            return self.get('class', class_id)(*args)
        elif kwargs:
            # unpack dictionary
            return self.get('class', class_id)(**kwargs)
        else:
            error_msg = 'Submit attributes as arguments or keyword arguments'
            raise TypeError(error_msg)



# -*- coding: utf-8 -*-
"""
chanjo.config.core
~~~~~~~~~~~~~~~~~~~

Works out of the box with any module that exposes the proper 'load' and
'dump' methods will work. Fox example 'toml' and 'yaml'.
"""
from __future__ import absolute_import
import io
import json

from path import path


class Config(dict):

    """Config object to store nested levels of config values.

    Reads values from a file and updates existing defaults. Can also save
    the updates key-value pairs back to a file.

    Customizable to use e.g. TOML, JSON, or YAML - it just requires an
    object with 'load' and 'dump' methods. Uses built in ``json`` module
    by default.

    Args:
        config_path (Optional[path]): location of the config file
        defaults (Optional[dict]): lowest priority default values
        markup (Optional[object]): markup object, default: json
        save_options (Optional[dict]): formatting when saving, e.g.
            ``{'indent': 4, 'sort_keys': True}``
    """

    def __init__(self, config_path=None, defaults=None, markup=json,
                 save_options=None):
        super(Config, self).__init__()
        self.user_data = {}
        self.save_options = (save_options or {})
        # trust the user knows what she's doing
        self.markup = markup
        # update the defaults with config data
        self.update((defaults or {}))

        self.config_path = path(config_path) if config_path else None
        if self.config_path and self.config_path.isfile():
            # Read data from possible config
            with io.open(self.config_path, encoding='utf-8') as handle:
                self.load(handle)

    def load(self, read_handle):
        try:
            values = self.markup.load(read_handle) or {}
            self.user_data.update(values)
            self.update(**self.user_data)
        except AttributeError:
            raise NotImplementedError("Markup (%s) must expose a 'load'-method"
                                      % str(self.markup))
        except ValueError as ex:
            raise ValueError("Bad syntax: %s" % ex)
        return self

    def save(self, **options):
        """Save the current key-value pairs using the write handle.

        Args:
            options (kwargs, optional): Options to pass to dump (like 'indent')

        Returns:
            Config: returns itself for chainability
        """
        options = (options or self.save_options)

        with self.config_path.open('w') as write_handle:
            self.markup.dump(self.user_data, write_handle, **options)
        return self

    def set(self, dot_key, value, scope=None):
        """Update a config key-value pair."""
        if scope is None:
            scope = self
        section, key = self._resolve_key(dot_key, base=scope)
        # Set key-value pair
        section[key] = value
        return self

    def _resolve_key(self, dot_key, base=None):
        """Resolve a "dot key" (e.g. person.age).

        Private method.

        Args:
            dot_key (str):
        """
        # The key can be provided as a path to nested pairs
        key_parts = dot_key.split('.')
        key_parts.reverse()

        if base is None:
            section = self
        else:
            section = base

        while len(key_parts) > 1:
            key_part = key_parts.pop()
            if key_part not in section:
                section[key_part] = {}

            section = section[key_part]

        return section, key_parts[0]
